"""Tests for `$ fio load`."""

from functools import partial
import json
import os
import shutil

import pytest

import fiona
from fiona.fio.main import main_group
from fiona.model import ObjectEncoder


def test_err(runner):
    result = runner.invoke(
        main_group, ['load'], '', catch_exceptions=False)
    assert result.exit_code == 2


def test_exception(tmpdir, runner):
    tmpfile = str(tmpdir.mkdir('tests').join('test_exception.shp'))
    result = runner.invoke(main_group, [
        'load', '-f', 'Shapefile', tmpfile
    ], '42', catch_exceptions=False)
    assert result.exit_code == 1


def test_collection(tmpdir, feature_collection, runner):
    tmpfile = str(tmpdir.mkdir('tests').join('test_collection.shp'))
    result = runner.invoke(
        main_group, ['load', '-f', 'Shapefile', tmpfile], feature_collection)
    assert result.exit_code == 0
    assert len(fiona.open(tmpfile)) == 2


def test_seq_rs(feature_seq_pp_rs, tmpdir, runner):
    tmpfile = str(tmpdir.mkdir('tests').join('test_seq_rs.shp'))
    result = runner.invoke(
        main_group, ['load', '-f', 'Shapefile', tmpfile], feature_seq_pp_rs)
    assert result.exit_code == 0
    assert len(fiona.open(tmpfile)) == 2


def test_seq_no_rs(tmpdir, runner, feature_seq):
    tmpfile = str(tmpdir.mkdir('tests').join('test_seq_no_rs.shp'))
    result = runner.invoke(main_group, [
        'load', '-f', 'Shapefile', tmpfile], feature_seq)
    assert result.exit_code == 0
    assert len(fiona.open(tmpfile)) == 2


def test_dst_crs_default_to_src_crs(tmpdir, runner, feature_seq):
    """When --dst-crs is not given default to --src-crs."""
    tmpfile = str(tmpdir.mkdir('tests').join('test_src_vs_dst_crs.shp'))
    result = runner.invoke(main_group, [
        'load',
        '--src-crs',
        'EPSG:32617',
        '-f', 'Shapefile',
        tmpfile
    ], feature_seq)
    assert result.exit_code == 0
    with fiona.open(tmpfile) as src:
        assert src.crs == {'init': 'epsg:32617'}
        assert len(src) == len(feature_seq.splitlines())


def test_different_crs(tmpdir, runner, feature_seq):
    tmpfile = str(tmpdir.mkdir('tests').join('test_different_crs.shp'))
    result = runner.invoke(
        main_group, [
            'load', '--src-crs', 'EPSG:32617', '--dst-crs', 'EPSG:32610',
            '-f', 'Shapefile', tmpfile
        ], feature_seq)
    assert result.exit_code == 0
    with fiona.open(tmpfile) as src:
        assert src.crs == {'init': 'epsg:32610'}
        assert len(src) == len(feature_seq.splitlines())


def test_dst_crs_no_src(tmpdir, runner, feature_seq):
    tmpfile = str(tmpdir.mkdir('tests').join('test_dst_crs_no_src.shp'))
    result = runner.invoke(main_group, [
        'load',
        '--dst-crs',
        'EPSG:32610',
        '-f', 'Shapefile',
        tmpfile
    ], feature_seq)
    assert result.exit_code == 0
    with fiona.open(tmpfile) as src:
        assert src.crs == {'init': 'epsg:32610'}
        assert len(src) == len(feature_seq.splitlines())


def test_fio_load_layer(tmpdir, runner):
    outdir = str(tmpdir.mkdir('tests').mkdir('test_fio_load_layer'))
    try:
        feature = {
            'type': 'Feature',
            'properties': {'key': 'value'},
            'geometry': {
                'type': 'Point',
                'coordinates': (5.0, 39.0)
            }
        }
        sequence = os.linesep.join(map(partial(json.dumps, cls=ObjectEncoder), [feature, feature]))
        result = runner.invoke(main_group, [
            'load',
            outdir,
            '--driver', 'ESRI Shapefile',
            '--src-crs', 'EPSG:4236',
            '--layer', 'test_layer'],
            input=sequence)
        assert result.exit_code == 0

        with fiona.open(outdir) as src:
            assert len(src) == 2
            assert src.name == 'test_layer'
            assert src.schema['geometry'] == 'Point'

    finally:
        shutil.rmtree(outdir)

@pytest.mark.iconv
def test_creation_options(tmpdir, runner, feature_seq):
    tmpfile = str(tmpdir.mkdir("tests").join("test.shp"))
    result = runner.invoke(
        main_group,
        ["load", "-f", "Shapefile", "--co", "ENCODING=LATIN1", tmpfile],
        feature_seq,
    )
    assert result.exit_code == 0
    assert tmpdir.join("tests/test.cpg").read() == "LATIN1"


@pytest.mark.parametrize("extension, driver", [
    ("shp", "ESRI Shapefile"),
    ("geojson", "GeoJSON"),
    ("json", "GeoJSON"),
    ("gpkg", "GPKG"),
    ("SHP", "ESRI Shapefile"),
])
def test_load__auto_detect_format(tmpdir, runner, feature_seq, extension, driver):
    tmpfile = str(tmpdir.mkdir('tests').join(f'test_src_vs_dst_crs.{extension}'))
    result = runner.invoke(main_group, [
        'load',
        '--src-crs',
        'EPSG:32617',
        tmpfile
    ], feature_seq)
    assert result.exit_code == 0
    with fiona.open(tmpfile.lower()) as src:
        assert src.crs == {'init': 'epsg:32617'}
        assert len(src) == len(feature_seq.splitlines())
        assert src.driver == driver


def test_fio_load_layer_append(tmpdir, runner):
    """Checking append mode."""
    outdir = str(tmpdir.mkdir("tests").mkdir("test_fio_load_layer"))
    try:
        feature = {
            "type": "Feature",
            "properties": {"key": "value"},
            "geometry": {"type": "Point", "coordinates": (5.0, 39.0)},
        }
        sequence = os.linesep.join(
            map(partial(json.dumps, cls=ObjectEncoder), [feature, feature])
        )

        # Write mode to create layer.
        result = runner.invoke(
            main_group,
            [
                "load",
                outdir,
                "--driver",
                "ESRI Shapefile",
                "--src-crs",
                "EPSG:4236",
                "--layer",
                "test_layer",
            ],
            input=sequence,
        )
        assert result.exit_code == 0

        # Here's the append.
        result = runner.invoke(
            main_group,
            [
                "load",
                outdir,
                "--driver=ESRI Shapefile",
                "--src-crs=EPSG:4236",
                "--layer=test_layer",
                "--append",
            ],
            input=sequence,
        )
        assert result.exit_code == 0

        with fiona.open(outdir) as src:
            assert len(src) == 4
            assert src.name == "test_layer"
            assert src.schema["geometry"] == "Point"

    finally:
        shutil.rmtree(outdir)



"""Tests of fiona.env"""

import os
import sys
from unittest import mock

import boto3
import pytest

import fiona
from fiona import _env
from fiona.env import getenv, hasenv, ensure_env, ensure_env_with_credentials
from fiona.errors import FionaDeprecationWarning
from fiona.session import AWSSession, GSSession


def test_nested_credentials(monkeypatch):
    """Check that rasterio.open() doesn't wipe out surrounding credentials"""

    @ensure_env_with_credentials
    def fake_opener(path):
        return fiona.env.getenv()

    with fiona.env.Env(
        session=AWSSession(aws_access_key_id="foo", aws_secret_access_key="bar")
    ):
        assert fiona.env.getenv()["AWS_ACCESS_KEY_ID"] == "foo"
        assert fiona.env.getenv()["AWS_SECRET_ACCESS_KEY"] == "bar"

        monkeypatch.setenv("AWS_ACCESS_KEY_ID", "lol")
        monkeypatch.setenv("AWS_SECRET_ACCESS_KEY", "wut")
        gdalenv = fake_opener("s3://foo/bar")
        assert gdalenv["AWS_ACCESS_KEY_ID"] == "foo"
        assert gdalenv["AWS_SECRET_ACCESS_KEY"] == "bar"


def test_ensure_env_decorator(gdalenv):
    @ensure_env
    def f():
        return getenv()["FIONA_ENV"]

    assert f() is True


def test_ensure_env_decorator_sets_gdal_data(gdalenv, monkeypatch):
    """fiona.env.ensure_env finds GDAL from environment"""

    @ensure_env
    def f():
        return getenv()["GDAL_DATA"]

    monkeypatch.setenv("GDAL_DATA", "/lol/wut")
    assert f() == "/lol/wut"


@mock.patch("fiona._env.GDALDataFinder.find_file")
def test_ensure_env_decorator_sets_gdal_data_prefix(
    find_file, gdalenv, monkeypatch, tmpdir
):
    """fiona.env.ensure_env finds GDAL data under a prefix"""

    @ensure_env
    def f():
        return getenv()["GDAL_DATA"]

    find_file.return_value = None
    tmpdir.ensure("share/gdal/header.dxf")
    monkeypatch.delenv("GDAL_DATA", raising=False)
    monkeypatch.setattr(_env, "__file__", str(tmpdir.join("fake.py")))
    monkeypatch.setattr(sys, "prefix", str(tmpdir))

    assert f() == str(tmpdir.join("share").join("gdal"))


@mock.patch("fiona._env.GDALDataFinder.find_file")
def test_ensure_env_decorator_sets_gdal_data_wheel(
    find_file, gdalenv, monkeypatch, tmpdir
):
    """fiona.env.ensure_env finds GDAL data in a wheel"""

    @ensure_env
    def f():
        return getenv()["GDAL_DATA"]

    find_file.return_value = None
    tmpdir.ensure("gdal_data/header.dxf")
    monkeypatch.delenv("GDAL_DATA", raising=False)
    monkeypatch.setattr(
        _env, "__file__", str(tmpdir.join(os.path.basename(_env.__file__)))
    )

    assert f() == str(tmpdir.join("gdal_data"))


@mock.patch("fiona._env.GDALDataFinder.find_file")
def test_ensure_env_with_decorator_sets_gdal_data_wheel(
    find_file, gdalenv, monkeypatch, tmpdir
):
    """fiona.env.ensure_env finds GDAL data in a wheel"""

    @ensure_env_with_credentials
    def f(*args):
        return getenv()["GDAL_DATA"]

    find_file.return_value = None
    tmpdir.ensure("gdal_data/header.dxf")
    monkeypatch.delenv("GDAL_DATA", raising=False)
    monkeypatch.setattr(
        _env, "__file__", str(tmpdir.join(os.path.basename(_env.__file__)))
    )

    assert f("foo") == str(tmpdir.join("gdal_data"))


def test_ensure_env_crs(path_coutwildrnp_shp):
    """Decoration of .crs works"""
    assert fiona.open(path_coutwildrnp_shp).crs


def test_env_default_env(path_coutwildrnp_shp):
    with fiona.open(path_coutwildrnp_shp):
        assert hasenv()


def test_nested_gs_credentials(monkeypatch):
    """Check that rasterio.open() doesn't wipe out surrounding credentials"""

    @ensure_env_with_credentials
    def fake_opener(path):
        return fiona.env.getenv()

    with fiona.env.Env(session=GSSession(google_application_credentials="foo")):
        assert fiona.env.getenv()["GOOGLE_APPLICATION_CREDENTIALS"] == "foo"

        gdalenv = fake_opener("gs://foo/bar")
        assert gdalenv["GOOGLE_APPLICATION_CREDENTIALS"] == "foo"


def test_aws_session(gdalenv):
    """Create an Env with a boto3 session."""
    aws_session = boto3.Session(
        aws_access_key_id="id",
        aws_secret_access_key="key",
        aws_session_token="token",
        region_name="null-island-1",
    )
    with pytest.warns(FionaDeprecationWarning):
        with fiona.env.Env(session=aws_session) as s:
            assert (
                s.session._session.get_credentials().get_frozen_credentials().access_key
                == "id"
            )
            assert (
                s.session._session.get_credentials().get_frozen_credentials().secret_key
                == "key"
            )
            assert (
                s.session._session.get_credentials().get_frozen_credentials().token
                == "token"
            )
            assert s.session._session.region_name == "null-island-1"



import os

import pytest
from click.testing import CliRunner

import fiona
from fiona.model import Feature
from fiona.fio.main import main_group


def create_sample_data(filename, driver, **extra_meta):
    meta = {"driver": driver, "schema": {"geometry": "Point", "properties": {}}}
    meta.update(extra_meta)
    with fiona.open(filename, "w", **meta) as dst:
        dst.write(
            Feature.from_dict(
                **{
                    "geometry": {
                        "type": "Point",
                        "coordinates": (0, 0),
                    },
                    "properties": {},
                }
            )
        )
    assert os.path.exists(filename)


drivers = ["ESRI Shapefile", "GeoJSON"]


@pytest.mark.parametrize("driver", drivers)
def test_remove(tmpdir, driver):
    extension = {"ESRI Shapefile": "shp", "GeoJSON": "json"}[driver]
    filename = f"delete_me.{extension}"
    filename = str(tmpdir.join(filename))
    create_sample_data(filename, driver)

    result = CliRunner().invoke(main_group, ["rm", filename, "--yes"])
    assert result.exit_code == 0
    assert not os.path.exists(filename)


has_gpkg = "GPKG" in fiona.supported_drivers.keys()


@pytest.mark.skipif(not has_gpkg, reason="Requires GPKG driver")
def test_remove_layer(tmpdir):
    filename = str(tmpdir.join("a_filename.gpkg"))
    create_sample_data(filename, "GPKG", layer="layer1")
    create_sample_data(filename, "GPKG", layer="layer2")
    assert fiona.listlayers(filename) == ["layer1", "layer2"]

    result = CliRunner().invoke(
        main_group, ["rm", filename, "--layer", "layer2", "--yes"]
    )
    assert result.exit_code == 0
    assert os.path.exists(filename)
    assert fiona.listlayers(filename) == ["layer1"]



"""$ fio env"""

import json
import os

import click

import fiona
from fiona._env import GDALDataFinder, PROJDataFinder


@click.command(short_help="Print information about the fio environment.")
@click.option('--formats', 'key', flag_value='formats', default=True,
              help="Enumerate the available formats.")
@click.option('--credentials', 'key', flag_value='credentials', default=False,
              help="Print credentials.")
@click.option('--gdal-data', 'key', flag_value='gdal_data', default=False,
              help="Print GDAL data path.")
@click.option('--proj-data', 'key', flag_value='proj_data', default=False,
              help="Print PROJ data path.")
@click.pass_context
def env(ctx, key):
    """Print information about the Fiona environment: available
    formats, etc.
    """
    stdout = click.get_text_stream('stdout')
    with ctx.obj['env'] as env:
        if key == 'formats':
            for k, v in sorted(fiona.supported_drivers.items()):
                modes = ', '.join("'" + m + "'" for m in v)
                stdout.write(f"{k} (modes {modes})\n")
            stdout.write('\n')
        elif key == 'credentials':
            click.echo(json.dumps(env.session.credentials))
        elif key == 'gdal_data':
            click.echo(os.environ.get('GDAL_DATA') or GDALDataFinder().search())
        elif key == 'proj_data':
            click.echo(os.environ.get('PROJ_DATA', os.environ.get('PROJ_LIB')) or PROJDataFinder().search())



# Cross-platform API functions.

cdef extern from "cpl_conv.h":
    void *  CPLMalloc (size_t)
    void    CPLFree (void *ptr)
    void    CPLSetThreadLocalConfigOption (char *key, char *val)
    const char *CPLGetConfigOption (char *, char *)

cdef extern from "cpl_vsi.h":
    ctypedef struct VSILFILE:
        pass
    int VSIFCloseL (VSILFILE *)
    VSILFILE * VSIFileFromMemBuffer (const char * filename,
                                     unsigned char * data,
                                     int data_len,
                                     int take_ownership)
    int VSIUnlink (const char * pathname)

ctypedef int OGRErr
ctypedef struct OGREnvelope:
    double MinX
    double MaxX
    double MinY
    double MaxY



# Fiona's date and time is founded on RFC 3339.
#
# OGR knows 3 time "zones": GMT, "local time", amd "unknown". Fiona, when
# writing will convert times with a timezone offset to GMT (Z) and otherwise
# will write times with the unknown zone.

import logging
import re

log = logging.getLogger("Fiona")


# Fiona's 'date', 'time', and 'datetime' types are sub types of 'str'.


class FionaDateType(str):
    """Dates without time."""


class FionaTimeType(str):
    """Times without dates."""


class FionaDateTimeType(str):
    """Dates and times."""


pattern_date = re.compile(r"(\d\d\d\d)(-)?(\d\d)(-)?(\d\d)")
pattern_time = re.compile(
    r"(\d\d)(:)?(\d\d)(:)?(\d\d)?(\.\d+)?(Z|([+-])?(\d\d)?(:)?(\d\d))?")
pattern_datetime = re.compile(
    r"(\d\d\d\d)(-)?(\d\d)(-)?(\d\d)(T)?(\d\d)(:)?(\d\d)(:)?(\d\d)?(\.\d+)?(Z|([+-])?(\d\d)?(:)?(\d\d))?")


class group_accessor:
    def __init__(self, m):
        self.match = m

    def group(self, i):
        try:
            return self.match.group(i) or 0
        except IndexError:
            return 0


def parse_time(text):
    """ Given a time, returns a datetime tuple

    Parameters
    ----------
    text: string to be parsed

    Returns
    -------
        (int, int , int, int, int, int, int, int):
            datetime tuple: (year, month, day, hour, minute, second, microsecond, utcoffset in minutes or None)

    """
    match = re.search(pattern_time, text)
    if match is None:
        raise ValueError("Time data '%s' does not match pattern" % text)
    g = group_accessor(match)
    log.debug("Match groups: %s", match.groups())

    if g.group(8) == '-':
        tz = -1.0 * (int(g.group(9)) * 60 + int(g.group(11)))
    elif g.group(8) == '+':
        tz = int(g.group(9)) * 60 + int(g.group(11))
    else:
        tz = None

    return (0, 0, 0,
            int(g.group(1)),
            int(g.group(3)),
            int(g.group(5)),
            int(1000000.0 * float(g.group(6))),
            tz
            )


def parse_date(text):
    """Given a date, returns a datetime tuple

    Parameters
    ----------
    text: string to be parsed

    Returns
    -------
        (int, int , int, int, int, int, int, int):
            datetime tuple: (year, month, day, hour, minute, second, microsecond, utcoffset in minutes or None)
    """
    match = re.search(pattern_date, text)
    if match is None:
        raise ValueError("Time data '%s' does not match pattern" % text)
    g = group_accessor(match)
    log.debug("Match groups: %s", match.groups())
    return (
        int(g.group(1)),
        int(g.group(3)),
        int(g.group(5)),
        0, 0, 0, 0, None)


def parse_datetime(text):
    """Given a datetime, returns a datetime tuple

    Parameters
    ----------
    text: string to be parsed

    Returns
    -------
        (int, int , int, int, int, int, int, int):
            datetime tuple: (year, month, day, hour, minute, second, microsecond, utcoffset in minutes or None)
    """
    match = re.search(pattern_datetime, text)
    if match is None:
        raise ValueError("Time data '%s' does not match pattern" % text)
    g = group_accessor(match)
    log.debug("Match groups: %s", match.groups())

    if g.group(14) == '-':
        tz = -1.0 * (int(g.group(15)) * 60 + int(g.group(17)))
    elif g.group(14) == '+':
        tz = int(g.group(15)) * 60 + int(g.group(17))
    else:
        tz = None

    return (
        int(g.group(1)),
        int(g.group(3)),
        int(g.group(5)),
        int(g.group(7)),
        int(g.group(9)),
        int(g.group(11)),
        int(1000000.0 * float(g.group(12))),
        tz)



"""Note well: collection slicing is deprecated!"""

import tempfile
import shutil
import os
from collections import OrderedDict
import pytest

from fiona.env import GDALVersion
import fiona
from fiona.drvsupport import supported_drivers, _driver_supports_mode
from fiona.errors import FionaDeprecationWarning
from fiona.model import Feature

from .conftest import get_temp_filename

gdal_version = GDALVersion.runtime()


def test_collection_get(path_coutwildrnp_shp):
    with fiona.open(path_coutwildrnp_shp) as src:
        result = src[5]
        assert result.id == "5"


def test_collection_slice(path_coutwildrnp_shp):
    with pytest.warns(FionaDeprecationWarning), fiona.open(path_coutwildrnp_shp) as src:
        results = src[:5]
        assert isinstance(results, list)
        assert len(results) == 5
        assert results[4].id == "4"


def test_collection_iterator_slice(path_coutwildrnp_shp):
    with fiona.open(path_coutwildrnp_shp) as src:
        results = list(src.items(5))
        assert len(results) == 5
        k, v = results[4]
        assert k == 4
        assert v.id == "4"


def test_collection_iterator_next(path_coutwildrnp_shp):
    with fiona.open(path_coutwildrnp_shp) as src:
        k, v = next(src.items(5, None))
        assert k == 5
        assert v.id == "5"


@pytest.fixture(
    scope="module",
    params=[
        driver
        for driver in supported_drivers
        if _driver_supports_mode(driver, "w")
        and driver not in {"DGN", "MapInfo File", "GPSTrackMaker", "GPX", "BNA", "DXF"}
    ],
)
def slice_dataset_path(request):
    """Create temporary datasets for test_collection_iterator_items_slice()"""

    driver = request.param
    min_id = 0
    max_id = 9

    def get_schema(driver):
        special_schemas = {
            "CSV": {"geometry": None, "properties": OrderedDict([("position", "int")])}
        }
        return special_schemas.get(
            driver,
            {"geometry": "Point", "properties": OrderedDict([("position", "int")])},
        )

    def get_records(driver, range):
        special_records1 = {
            "CSV": [
                Feature.from_dict(**{"geometry": None, "properties": {"position": i}})
                for i in range
            ],
            "PCIDSK": [
                Feature.from_dict(
                    **{
                        "geometry": {
                            "type": "Point",
                            "coordinates": (0.0, float(i), 0.0),
                        },
                        "properties": {"position": i},
                    }
                )
                for i in range
            ],
        }
        return special_records1.get(
            driver,
            [
                Feature.from_dict(
                    **{
                        "geometry": {"type": "Point", "coordinates": (0.0, float(i))},
                        "properties": {"position": i},
                    }
                )
                for i in range
            ],
        )

    schema = get_schema(driver)
    records = get_records(driver, range(min_id, max_id + 1))

    create_kwargs = {}
    if driver == "FlatGeobuf":
        create_kwargs["SPATIAL_INDEX"] = False

    tmpdir = tempfile.mkdtemp()
    path = os.path.join(tmpdir, get_temp_filename(driver))

    with fiona.open(path, "w", driver=driver, schema=schema, **create_kwargs) as c:
        c.writerecords(records)
    yield path
    shutil.rmtree(tmpdir)


@pytest.mark.parametrize(
    "args",
    [
        (0, 5, None),
        (1, 5, None),
        (-5, None, None),
        (-5, -1, None),
        (0, None, None),
        (5, None, None),
        (8, None, None),
        (9, None, None),
        (10, None, None),
        (0, 5, 2),
        (0, 5, 2),
        (1, 5, 2),
        (-5, None, 2),
        (-5, -1, 2),
        (0, None, 2),
        (0, 8, 2),
        (0, 9, 2),
        (0, 10, 2),
        (1, 8, 2),
        (1, 9, 2),
        (1, 10, 2),
        (1, None, 2),
        (5, None, 2),
        (5, None, -1),
        (5, None, -2),
        (5, None, None),
        (4, None, -2),
        (-1, -5, -1),
        (-5, None, -1),
        (0, 5, 1),
        (5, 15, 1),
        (15, 30, 1),
        (5, 0, -1),
        (15, 5, -1),
        (30, 15, -1),
        (0, 5, 2),
        (5, 15, 2),
        (15, 30, 2),
        (5, 0, -2),
        (15, 5, -2),
        (30, 15, -2),
    ],
)
@pytest.mark.filterwarnings("ignore:.*OLC_FASTFEATURECOUNT*")
@pytest.mark.filterwarnings("ignore:.*OLCFastSetNextByIndex*")
def test_collection_iterator_items_slice(slice_dataset_path, args):
    """Test if c.items(start, stop, step) returns the correct features."""

    start, stop, step = args
    min_id = 0
    max_id = 9

    positions = list(range(min_id, max_id + 1))[start:stop:step]

    with fiona.open(slice_dataset_path, "r") as c:
        items = list(c.items(start, stop, step))
        assert len(items) == len(positions)
        record_positions = [int(item[1]["properties"]["position"]) for item in items]
        for expected_position, record_position in zip(positions, record_positions):
            assert expected_position == record_position


def test_collection_iterator_keys_next(path_coutwildrnp_shp):
    with fiona.open(path_coutwildrnp_shp) as src:
        k = next(src.keys(5, None))
        assert k == 5



import logging
import sys
import os
import itertools
from .conftest import requires_gpkg

import pytest

import fiona
from fiona.errors import DatasetDeleteError
from fiona.model import Feature


def create_sample_data(filename, driver, **extra_meta):
    meta = {"driver": driver, "schema": {"geometry": "Point", "properties": {}}}
    meta.update(extra_meta)
    with fiona.open(filename, "w", **meta) as dst:
        dst.write(
            Feature.from_dict(
                **{
                    "geometry": {
                        "type": "Point",
                        "coordinates": (0, 0),
                    },
                    "properties": {},
                }
            )
        )
    assert os.path.exists(filename)


drivers = ["ESRI Shapefile", "GeoJSON"]
kinds = ["path", "collection"]
specify_drivers = [True, False]
test_data = itertools.product(drivers, kinds, specify_drivers)


@pytest.mark.parametrize("driver, kind, specify_driver", test_data)
def test_remove(tmpdir, kind, driver, specify_driver):
    """Test various dataset removal operations"""
    extension = {"ESRI Shapefile": "shp", "GeoJSON": "json"}[driver]
    filename = f"delete_me.{extension}"
    output_filename = str(tmpdir.join(filename))

    create_sample_data(output_filename, driver=driver)
    if kind == "collection":
        to_delete = fiona.open(output_filename, "r")
    else:
        to_delete = output_filename

    assert os.path.exists(output_filename)
    if specify_driver:
        fiona.remove(to_delete, driver=driver)
    else:
        fiona.remove(to_delete)
    assert not os.path.exists(output_filename)


def test_remove_nonexistent(tmpdir):
    """Attempting to remove a file that does not exist results in an OSError"""
    filename = str(tmpdir.join("does_not_exist.shp"))
    assert not os.path.exists(filename)
    with pytest.raises(OSError):
        fiona.remove(filename)


@requires_gpkg
def test_remove_layer(tmpdir):
    filename = str(tmpdir.join("a_filename.gpkg"))
    create_sample_data(filename, "GPKG", layer="layer1")
    create_sample_data(filename, "GPKG", layer="layer2")
    create_sample_data(filename, "GPKG", layer="layer3")
    create_sample_data(filename, "GPKG", layer="layer4")
    assert fiona.listlayers(filename) == ["layer1", "layer2", "layer3", "layer4"]

    # remove by index
    fiona.remove(filename, layer=2)
    assert fiona.listlayers(filename) == ["layer1", "layer2", "layer4"]

    # remove by name
    fiona.remove(filename, layer="layer2")
    assert fiona.listlayers(filename) == ["layer1", "layer4"]

    # remove by negative index
    fiona.remove(filename, layer=-1)
    assert fiona.listlayers(filename) == ["layer1"]

    # invalid layer name
    with pytest.raises(ValueError):
        fiona.remove(filename, layer="invalid_layer_name")

    # invalid layer index
    with pytest.raises(DatasetDeleteError):
        fiona.remove(filename, layer=999)


def test_remove_layer_shapefile(tmpdir):
    """Removal of layer in shapefile actually deletes the datasource"""
    filename = str(tmpdir.join("a_filename.shp"))
    create_sample_data(filename, "ESRI Shapefile")
    fiona.remove(filename, layer=0)
    assert not os.path.exists(filename)


def test_remove_layer_geojson(tmpdir):
    """Removal of layers is not supported by GeoJSON driver

    The reason for failure is slightly different between GDAL 2.2+ and < 2.2.
    With < 2.2 the datasource will fail to open in write mode (OSError), while
    with 2.2+ the datasource will open but the removal operation will fail (not
    supported).
    """
    filename = str(tmpdir.join("a_filename.geojson"))
    create_sample_data(filename, "GeoJSON")
    with pytest.raises((RuntimeError, OSError)):
        fiona.remove(filename, layer=0)
    assert os.path.exists(filename)
